@@ -103,6 +103,10 @@ typedef enum {
 #define M_PI 3.14159265358979323846
 #endif
 
+#ifndef M_SQRT1_2
+#define M_SQRT1_2  0.70710678118654752440  /* 1/sqrt(2) */
+#endif
+
 /*****************************************************************************
  * Forward declaration
  ****************************************************************************/
@@ -117,7 +121,8 @@ static Bool ConvertProc(LocalDevicePtr, int, int, int, int, int, int, int, int,
 static Bool DeviceInit(DeviceIntPtr);
 static Bool DeviceOn(DeviceIntPtr);
 static Bool DeviceOff(DeviceIntPtr);
-static Bool DeviceInit(DeviceIntPtr);
+static Bool DeviceClose(DeviceIntPtr);
+static Bool QueryHardware(LocalDevicePtr);
 
 
 InputDriverRec SYNAPTICS = {
@@ -199,6 +204,64 @@ SetDeviceAndProtocol(LocalDevicePtr local)
     }
 }
 
+/*
+ * Allocate and initialize memory for the SynapticsSHM data to hold driver
+ * parameter settings.
+ * The function will allocate shared memory if priv->shm_config is TRUE.
+ * The allocated data is initialized from priv->synpara_default.
+ */
+static Bool
+alloc_param_data(LocalDevicePtr local)
+{
+    int shmid;
+    SynapticsPrivate *priv = local->private;
+
+    if (priv->synpara)
+	return TRUE;			    /* Already allocated */
+
+    if (priv->shm_config) {
+	if ((shmid = xf86shmget(SHM_SYNAPTICS, 0, 0)) != -1)
+	    xf86shmctl(shmid, XF86IPC_RMID, NULL);
+	if ((shmid = xf86shmget(SHM_SYNAPTICS, sizeof(SynapticsSHM),
+				0777 | XF86IPC_CREAT)) == -1) {
+	    xf86Msg(X_ERROR, "%s error shmget\n", local->name);
+	    return FALSE;
+	}
+	if ((priv->synpara = (SynapticsSHM*)xf86shmat(shmid, NULL, 0)) == NULL) {
+	    xf86Msg(X_ERROR, "%s error shmat\n", local->name);
+	    return FALSE;
+	}
+    } else {
+	priv->synpara = xcalloc(1, sizeof(SynapticsSHM));
+	if (!priv->synpara)
+	    return FALSE;
+    }
+
+    *(priv->synpara) = priv->synpara_default;
+    return TRUE;
+}
+
+/*
+ * Free SynapticsSHM data previously allocated by alloc_param_data().
+ */
+static void
+free_param_data(SynapticsPrivate *priv)
+{
+    int shmid;
+
+    if (!priv->synpara)
+	return;
+
+    if (priv->shm_config) {
+	if ((shmid = xf86shmget(SHM_SYNAPTICS, 0, 0)) != -1)
+	    xf86shmctl(shmid, XF86IPC_RMID, NULL);
+    } else {
+	xfree(priv->synpara);
+    }
+
+    priv->synpara = NULL;
+}
+
 /*
  *  called by the module loader for initialization
  */
@@ -213,12 +276,11 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pointer optList;
 #endif
     char *str_par;
-    int shmid;
     unsigned long now;
     SynapticsSHM *pars;
     char *repeater;
 
-    /* allocate memory for SynaticsPrivateRec */
+    /* allocate memory for SynapticsPrivateRec */
     priv = xcalloc(1, sizeof(SynapticsPrivate));
     if (!priv)
 	return NULL;
@@ -277,28 +339,10 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     priv->touch_on.millis = now;
 
     /* install shared memory or normal memory for parameters */
-    priv->shm_config = FALSE;
-    if (xf86SetBoolOption(local->options, "SHMConfig", FALSE)) {
-	if ((shmid = xf86shmget(SHM_SYNAPTICS, 0, 0)) != -1)
-	    xf86shmctl(shmid, XF86IPC_RMID, NULL);
-	if ((shmid = xf86shmget(SHM_SYNAPTICS, sizeof(SynapticsSHM),
-				0777 | XF86IPC_CREAT)) == -1) {
-	    xf86Msg(X_ERROR, "%s error shmget\n", local->name);
-	    goto SetupProc_fail;
-	}
-	else if ((priv->synpara = (SynapticsSHM*) xf86shmat(shmid, NULL, 0)) == NULL) {
-	    xf86Msg(X_ERROR, "%s error shmat\n", local->name);
-	    goto SetupProc_fail;
-	}
-	priv->shm_config = TRUE;
-    } else {
-	priv->synpara = xcalloc(1, sizeof(SynapticsSHM));
-	if (!priv->synpara)
-	    goto SetupProc_fail;
-    }
+    priv->shm_config = xf86SetBoolOption(local->options, "SHMConfig", FALSE);
 
     /* read the parameters */
-    pars = priv->synpara;
+    pars = &priv->synpara_default;
     pars->left_edge = xf86SetIntOption(local->options, "LeftEdge", 1900);
     pars->right_edge = xf86SetIntOption(local->options, "RightEdge", 5400);
     pars->top_edge = xf86SetIntOption(local->options, "TopEdge", 1900);
@@ -309,6 +353,7 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pars->tap_move = xf86SetIntOption(local->options, "MaxTapMove", 220);
     pars->tap_time_2 = xf86SetIntOption(local->options, "MaxDoubleTapTime", 180);
     pars->click_time = xf86SetIntOption(local->options, "ClickTime", 100);
+    pars->fast_taps = xf86SetIntOption(local->options, "FastTaps", FALSE);
     pars->emulate_mid_button_time = xf86SetIntOption(local->options,
 							      "EmulateMidButtonTime", 75);
     pars->scroll_dist_vert = xf86SetIntOption(local->options, "VertScrollDelta", 100);
@@ -316,11 +361,11 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pars->edge_motion_min_z = xf86SetIntOption(local->options, "EdgeMotionMinZ", 30);
     pars->edge_motion_max_z = xf86SetIntOption(local->options, "EdgeMotionMaxZ", 160);
     pars->edge_motion_min_speed = xf86SetIntOption(local->options, "EdgeMotionMinSpeed", 1);
-    pars->edge_motion_max_speed = xf86SetIntOption(local->options, "EdgeMotionMaxSpeed", 200);
+    pars->edge_motion_max_speed = xf86SetIntOption(local->options, "EdgeMotionMaxSpeed", 400);
     pars->edge_motion_use_always = xf86SetBoolOption(local->options, "EdgeMotionUseAlways", FALSE);
     repeater = xf86SetStrOption(local->options, "Repeater", NULL);
     pars->updown_button_scrolling = xf86SetBoolOption(local->options, "UpDownScrolling", TRUE);
-    pars->touchpad_off = xf86SetBoolOption(local->options, "TouchpadOff", FALSE);
+    pars->touchpad_off = xf86SetIntOption(local->options, "TouchpadOff", 0);
     pars->guestmouse_off = xf86SetBoolOption(local->options, "GuestMouseOff", FALSE);
     pars->locked_drags = xf86SetBoolOption(local->options, "LockedDrags", FALSE);
     pars->tap_action[RT_TAP] = xf86SetIntOption(local->options, "RTCornerButton", 2);
@@ -332,10 +377,14 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     pars->tap_action[F3_TAP] = xf86SetIntOption(local->options, "TapButton3",     3);
     pars->circular_scrolling = xf86SetBoolOption(local->options, "CircularScrolling", FALSE);
     pars->circular_trigger   = xf86SetIntOption(local->options, "CircScrollTrigger", 0);
+    pars->circular_pad       = xf86SetBoolOption(local->options, "CircularPad", FALSE);
+    pars->palm_detect        = xf86SetBoolOption(local->options, "PalmDetect", TRUE);
+    pars->palm_min_width     = xf86SetIntOption(local->options, "PalmMinWidth", 10);
+    pars->palm_min_z         = xf86SetIntOption(local->options, "PalmMinZ", 200);
 
     str_par = xf86FindOptionValue(local->options, "MinSpeed");
     if ((!str_par) || (xf86sscanf(str_par, "%lf", &pars->min_speed) != 1))
-	pars->min_speed=0.02;
+	pars->min_speed=0.09;
     str_par = xf86FindOptionValue(local->options, "MaxSpeed");
     if ((!str_par) || (xf86sscanf(str_par, "%lf", &pars->max_speed) != 1))
 	pars->max_speed=0.18;
@@ -345,6 +394,9 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     str_par = xf86FindOptionValue(local->options, "CircScrollDelta");
     if ((!str_par) || (xf86sscanf(str_par, "%lf", &pars->scroll_dist_circ) != 1))
 	pars->scroll_dist_circ = 0.1;
+    str_par = xf86FindOptionValue(local->options, "CoastingSpeed");
+    if ((!str_par) || (xf86sscanf(str_par, "%lf", &pars->coasting_speed) != 1))
+	pars->coasting_speed = 0.0;
 
     if (pars->circular_trigger < 0 || pars->circular_trigger > 8) {
 	xf86Msg(X_WARNING, "Unknown circular scrolling trigger, using 0 (edges)");
@@ -362,6 +414,9 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 
     priv->largest_valid_x = MIN(pars->right_edge, XMAX_NOMINAL);
 
+    if (!alloc_param_data(local))
+	goto SetupProc_fail;
+
     priv->comm.buffer = XisbNew(local->fd, 200);
     DBG(9, XisbTrace(priv->comm.buffer, 1));
 
@@ -392,7 +447,6 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
     local->flags |= XI86_CONFIGURED;
 
     if (local->fd != -1) {
-	xf86RemoveEnabledDevice(local);
 	if (priv->comm.buffer) {
 	    XisbFree(priv->comm.buffer);
 	    priv->comm.buffer = NULL;
@@ -404,21 +458,13 @@ SynapticsPreInit(InputDriverPtr drv, IDevPtr dev, int flags)
 
  SetupProc_fail:
     if (local->fd >= 0) {
-	RemoveEnabledDevice(local->fd);
 	xf86CloseSerial(local->fd);
 	local->fd = -1;
     }
 
     if (priv->comm.buffer)
 	XisbFree(priv->comm.buffer);
-    if (priv->synpara) {
-	if (priv->shm_config) {
-	    if ((shmid = xf86shmget(SHM_SYNAPTICS, 0, 0)) != -1)
-		xf86shmctl(shmid, XF86IPC_RMID, NULL);
-	} else {
-	    xfree(priv->synpara);
-	}
-    }
+    free_param_data(priv);
     /* Freeing priv makes the X server crash. Don't know why.
        xfree(priv);
     */
@@ -450,25 +496,16 @@ DeviceControl(DeviceIntPtr dev, int mode)
 
     switch (mode) {
     case DEVICE_INIT:
-	DeviceInit(dev);
-	RetValue = Success;
+	RetValue = DeviceInit(dev);
 	break;
     case DEVICE_ON:
-	RetValue = DeviceOn( dev );
+	RetValue = DeviceOn(dev);
 	break;
     case DEVICE_OFF:
-	RetValue = DeviceOff( dev );
+	RetValue = DeviceOff(dev);
 	break;
     case DEVICE_CLOSE:
-	{
-	    int shmid;
-	    LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
-	    SynapticsPrivate *priv = (SynapticsPrivate *) (local->private);
-	    RetValue = DeviceOff( dev );
-	    if (priv->shm_config)
-		if ((shmid = xf86shmget(SHM_SYNAPTICS, 0, 0)) != -1)
-		    xf86shmctl(shmid, XF86IPC_RMID, NULL);
-	}
+	RetValue = DeviceClose(dev);
 	break;
     default:
 	RetValue = BadValue;
@@ -519,6 +556,8 @@ DeviceOff(DeviceIntPtr dev)
     DBG(3, ErrorF("Synaptics DeviceOff called\n"));
 
     if (local->fd != -1) {
+	TimerFree(priv->timer);
+	priv->timer = NULL;
 	xf86RemoveEnabledDevice(local);
 	priv->proto_ops->DeviceOffHook(local);
 	if (priv->comm.buffer) {
@@ -531,18 +570,34 @@ DeviceOff(DeviceIntPtr dev)
     return Success;
 }
 
+static Bool
+DeviceClose(DeviceIntPtr dev)
+{
+    Bool RetValue;
+    LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
+    SynapticsPrivate *priv = (SynapticsPrivate *) local->private;
+
+    RetValue = DeviceOff(dev);
+    free_param_data(priv);
+    return RetValue;
+}
+
 static Bool
 DeviceInit(DeviceIntPtr dev)
 {
     LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
-    unsigned char map[] = {0, 1, 2, 3, 4, 5, 6, 7};
+    unsigned char map[SYN_MAX_BUTTONS + 1];
+    int i;
 
     DBG(3, ErrorF("Synaptics DeviceInit called\n"));
 
+    for (i = 0; i <= SYN_MAX_BUTTONS; i++)
+	map[i] = i;
+
     dev->public.on = FALSE;
 
     InitPointerDeviceStruct((DevicePtr)dev, map,
-			    7,
+			    SYN_MAX_BUTTONS,
 			    miPointerGetMotionEvents, SynapticsCtrl,
 			    miPointerGetMotionBufferSize());
 
@@ -555,6 +610,9 @@ DeviceInit(DeviceIntPtr dev)
 
     xf86MotionHistoryAllocate(local);
 
+    if (!alloc_param_data(local))
+	return !Success;
+
     return Success;
 }
 
@@ -564,6 +622,31 @@ move_distance(int dx, int dy)
     return xf86sqrt((dx * dx) + (dy * dy));
 }
 
+/*
+ * Convert from absolute X/Y coordinates to a coordinate system where
+ * -1 corresponds to the left/upper edge and +1 corresponds to the
+ * right/lower edge.
+ */
+static void
+relative_coords(SynapticsPrivate *priv, int x, int y,
+		double *relX, double *relY)
+{
+    int minX = priv->synpara->left_edge;
+    int maxX = priv->synpara->right_edge;
+    int minY = priv->synpara->top_edge;
+    int maxY = priv->synpara->bottom_edge;
+    double xCenter = (minX + maxX) / 2.0;
+    double yCenter = (minY + maxY) / 2.0;
+
+    if ((maxX - xCenter > 0) && (maxY - yCenter > 0)) {
+	*relX = (x - xCenter) / (maxX - xCenter);
+	*relY = (y - yCenter) / (maxY - yCenter);
+    } else {
+	*relX = 0;
+	*relY = 0;
+    }
+}
+
 /* return angle of point relative to center */
 static double
 angle(SynapticsPrivate *priv, int x, int y)
@@ -586,11 +669,39 @@ diffa(double a1, double a2)
     return da;
 }
 
+static edge_type
+circular_edge_detection(SynapticsPrivate *priv, int x, int y)
+{
+    edge_type edge = 0;
+    double relX, relY, relR;
+
+    relative_coords(priv, x, y, &relX, &relY);
+    relR = relX * relX + relY * relY;
+
+    if (relR > 1) {
+	/* we are outside the ellipse enclosed by the edge parameters */
+	if (relX > M_SQRT1_2)
+	    edge |= RIGHT_EDGE;
+	else if (relX < -M_SQRT1_2)
+	    edge |= LEFT_EDGE;
+
+	if (relY < -M_SQRT1_2)
+	    edge |= TOP_EDGE;
+	else if (relY > M_SQRT1_2)
+	    edge |= BOTTOM_EDGE;
+    }
+
+    return edge;
+}
+
 static edge_type
 edge_detection(SynapticsPrivate *priv, int x, int y)
 {
     edge_type edge = 0;
 
+    if (priv->synpara->circular_pad)
+	return circular_edge_detection(priv, x, y);
+
     if (x > priv->synpara->right_edge)
 	edge |= RIGHT_EDGE;
     else if (x < priv->synpara->left_edge)
@@ -637,7 +748,8 @@ timerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
     return 0;
 }
 
-static int clamp(int val, int min, int max)
+static int
+clamp(int val, int min, int max)
 {
     if (val < min)
 	return min;
@@ -771,9 +883,12 @@ SynapticsDetectFinger(SynapticsPrivate *priv, struct SynapticsHwState *hw)
     finger = (((hw->z > para->finger_high) && !priv->finger_flag) ||
 	      ((hw->z > para->finger_low)  &&  priv->finger_flag));
 
+    if (!para->palm_detect)
+        return finger;
+
     /* palm detection */
     if (finger) {
-	if ((hw->z > 200) && (hw->fingerWidth > 10))
+	if ((hw->z > para->palm_min_z) && (hw->fingerWidth > para->palm_min_width))
 	    priv->palm = TRUE;
     } else {
 	priv->palm = FALSE;
@@ -794,9 +909,9 @@ SynapticsDetectFinger(SynapticsPrivate *priv, struct SynapticsHwState *hw)
 	    finger = FALSE;
 	else if (hw->z < priv->prev_z - 5)	/* z not stable, may be a palm */
 	    finger = FALSE;
-	else if (hw->z > 200)			/* z too large -> probably palm */
+	else if (hw->z > para->palm_min_z)	/* z too large -> probably palm */
 	    finger = FALSE;
-	else if (hw->fingerWidth > 10)		/* finger width too large -> probably palm */
+	else if (hw->fingerWidth > para->palm_min_width) /* finger width too large -> probably palm */
 	    finger = FALSE;
     }
     priv->prev_z = hw->z;
@@ -812,6 +927,11 @@ SelectTapButton(SynapticsPrivate *priv, edge_type edge)
 {
     TapEvent tap;
 
+    if (priv->synpara->touchpad_off == 2) {
+	priv->tap_button = 0;
+	return;
+    }    
+
     switch (priv->tap_max_fingers) {
     case 1:
     default:
@@ -849,12 +969,13 @@ SelectTapButton(SynapticsPrivate *priv, edge_type edge)
     }
 
     priv->tap_button = priv->synpara->tap_action[tap];
-    priv->tap_button = clamp(priv->tap_button, 0, 7);
+    priv->tap_button = clamp(priv->tap_button, 0, SYN_MAX_BUTTONS);
 }
 
 static void
 SetTapState(SynapticsPrivate *priv, enum TapState tap_state, int millis)
 {
+    SynapticsSHM *para = priv->synpara;
     DBG(7, ErrorF("SetTapState - %d -> %d (millis:%d)\n", priv->tap_state, tap_state, millis));
     switch (tap_state) {
     case TS_START:
@@ -862,7 +983,14 @@ SetTapState(SynapticsPrivate *priv, enum TapState tap_state, int millis)
 	priv->tap_max_fingers = 0;
 	break;
     case TS_1:
+	priv->tap_button_state = TBS_BUTTON_UP;
+	break;
     case TS_2A:
+	if (para->fast_taps)
+	    priv->tap_button_state = TBS_BUTTON_DOWN;
+	else 
+	    priv->tap_button_state = TBS_BUTTON_UP;
+	break;
     case TS_2B:
 	priv->tap_button_state = TBS_BUTTON_UP;
 	break;
@@ -870,7 +998,10 @@ SetTapState(SynapticsPrivate *priv, enum TapState tap_state, int millis)
 	priv->tap_button_state = TBS_BUTTON_DOWN;
 	break;
     case TS_SINGLETAP:
-	priv->tap_button_state = TBS_BUTTON_DOWN;
+	if (para->fast_taps)
+	    priv->tap_button_state = TBS_BUTTON_UP;
+	else 
+	    priv->tap_button_state = TBS_BUTTON_DOWN;
 	priv->touch_on.millis = millis;
 	break;
     default:
@@ -1005,14 +1136,37 @@ HandleTapProcessing(SynapticsPrivate *priv, struct SynapticsHwState *hw,
     return delay;
 }
 
-#define MOVE_HIST(a) (priv->move_hist[((priv->count_packet_finger-(a))%SYNAPTICS_MOVE_HISTORY)])
+#define HIST(a) (priv->move_hist[((priv->hist_index - (a) + SYNAPTICS_MOVE_HISTORY) % SYNAPTICS_MOVE_HISTORY)])
 
-static long ComputeDeltas(SynapticsPrivate *priv, struct SynapticsHwState *hw,
-			  edge_type edge, int *dxP, int *dyP)
+static void
+store_history(SynapticsPrivate *priv, int x, int y, unsigned int millis)
+{
+    int idx = (priv->hist_index + 1) % SYNAPTICS_MOVE_HISTORY;
+    priv->move_hist[idx].x = x;
+    priv->move_hist[idx].y = y;
+    priv->move_hist[idx].millis = millis;
+    priv->hist_index = idx;
+}
+
+/*
+ * Estimate the slope for the data sequence [x3, x2, x1, x0] by using
+ * linear regression to fit a line to the data and use the slope of the
+ * line.
+ */
+static double
+estimate_delta(double x0, double x1, double x2, double x3)
+{
+    return x0 * 0.3 + x1 * 0.1 - x2 * 0.1 - x3 * 0.3;
+}
+
+static long
+ComputeDeltas(SynapticsPrivate *priv, struct SynapticsHwState *hw,
+	      edge_type edge, int *dxP, int *dyP)
 {
     SynapticsSHM *para = priv->synpara;
     Bool moving_state;
-    int dist, dx, dy;
+    int dist;
+    double dx, dy;
     double speed, integral;
     long delay = 1000000000;
 
@@ -1037,8 +1191,12 @@ static long ComputeDeltas(SynapticsPrivate *priv, struct SynapticsHwState *hw,
 	!priv->vert_scroll_on && !priv->horiz_scroll_on && !priv->circ_scroll_on) {
 	delay = MIN(delay, 13);
 	if (priv->count_packet_finger > 3) { /* min. 3 packets */
-	    dx = (hw->x - MOVE_HIST(2).x) / 2;
-	    dy = (hw->y - MOVE_HIST(2).y) / 2;
+	    double tmpf;
+	    int x_edge_speed = 0;
+	    int y_edge_speed = 0;
+	    double dtime = (hw->millis - HIST(0).millis) / 1000.0;
+	    dx = estimate_delta(hw->x, HIST(0).x, HIST(1).x, HIST(2).x);
+	    dy = estimate_delta(hw->y, HIST(0).y, HIST(1).y, HIST(2).y);
 
 	    if ((priv->tap_state == TS_DRAG) || para->edge_motion_use_always) {
 		int minZ = para->edge_motion_min_z;
@@ -1054,15 +1212,25 @@ static long ComputeDeltas(SynapticsPrivate *priv, struct SynapticsHwState *hw,
 		} else {
 		    edge_speed = minSpd + (hw->z - minZ) * (maxSpd - minSpd) / (maxZ - minZ);
 		}
-		if (edge & RIGHT_EDGE) {
-		    dx += clamp(edge_speed - dx, 0, edge_speed);
-		} else if (edge & LEFT_EDGE) {
-		    dx -= clamp(edge_speed + dx, 0, edge_speed);
-		}
-		if (edge & TOP_EDGE) {
-		    dy -= clamp(edge_speed + dy, 0, edge_speed);
-		} else if (edge & BOTTOM_EDGE) {
-		    dy += clamp(edge_speed - dy, 0, edge_speed);
+		if (!priv->synpara->circular_pad) {
+		    /* on rectangular pad */
+		    if (edge & RIGHT_EDGE) {
+			x_edge_speed = edge_speed;
+		    } else if (edge & LEFT_EDGE) {
+			x_edge_speed = -edge_speed;
+		    }
+		    if (edge & TOP_EDGE) {
+			y_edge_speed = -edge_speed;
+		    } else if (edge & BOTTOM_EDGE) {
+			y_edge_speed = edge_speed;
+		    }
+		} else if (edge) {
+		    /* at edge of circular pad */
+		    double relX, relY;
+
+		    relative_coords(priv, hw->x, hw->y, &relX, &relY);
+		    x_edge_speed = (int)(edge_speed * relX);
+		    y_edge_speed = (int)(edge_speed * relY);
 		}
 	    }
 
@@ -1075,10 +1243,12 @@ static long ComputeDeltas(SynapticsPrivate *priv, struct SynapticsHwState *hw,
 		speed = para->min_speed;
 	    }
 
-	    /* save the fraction for adding to the next priv->count_packet */
-	    priv->frac_x = xf86modf((dx * speed) + priv->frac_x, &integral);
+	    /* save the fraction, report the integer part */
+	    tmpf = dx * speed + x_edge_speed * dtime + priv->frac_x;
+	    priv->frac_x = xf86modf(tmpf, &integral);
 	    dx = integral;
-	    priv->frac_y = xf86modf((dy * speed) + priv->frac_y, &integral);
+	    tmpf = dy * speed + y_edge_speed * dtime + priv->frac_y;
+	    priv->frac_y = xf86modf(tmpf, &integral);
 	    dy = integral;
 	}
 
@@ -1097,8 +1267,7 @@ static long ComputeDeltas(SynapticsPrivate *priv, struct SynapticsHwState *hw,
     *dyP = dy;
 
     /* generate a history of the absolute positions */
-    MOVE_HIST(0).x = hw->x;
-    MOVE_HIST(0).y = hw->y;
+    store_history(priv, hw->x, hw->y, hw->millis);
 
     return delay;
 }
@@ -1108,15 +1277,56 @@ struct ScrollData {
 };
 
 static void
+start_coasting(SynapticsPrivate *priv, struct SynapticsHwState *hw, edge_type edge)
+{
+    SynapticsSHM *para = priv->synpara;
+
+    priv->autoscroll_y = 0.0;
+    priv->autoscroll_x = 0.0;
+
+    if ((priv->scroll_packet_count > 3) && (para->coasting_speed > 0.0)) {
+	double pkt_time = (HIST(0).millis - HIST(3).millis) / 1000.0;
+	if (priv->vert_scroll_on ||
+	    (priv->circ_scroll_on && priv->circ_scroll_vert)) {
+	    double dy = estimate_delta(HIST(0).y, HIST(1).y, HIST(2).y, HIST(3).y);
+	    int sdelta = para->scroll_dist_vert;
+	    if ((edge & RIGHT_EDGE) && pkt_time > 0 && sdelta > 0) {
+		double scrolls_per_sec = dy / pkt_time / sdelta;
+		if (fabs(scrolls_per_sec) >= para->coasting_speed) {
+		    priv->autoscroll_yspd = scrolls_per_sec;
+		    priv->autoscroll_y = (hw->y - priv->scroll_y) / (double)sdelta;
+		}
+	    }
+	} else {
+	    double dx = estimate_delta(HIST(0).x, HIST(1).x, HIST(2).x, HIST(3).x);
+	    int sdelta = para->scroll_dist_horiz;
+	    if ((edge & BOTTOM_EDGE) && pkt_time > 0 && sdelta > 0) {
+		double scrolls_per_sec = dx / pkt_time / sdelta;
+		if (fabs(scrolls_per_sec) >= para->coasting_speed) {
+		    priv->autoscroll_xspd = scrolls_per_sec;
+		    priv->autoscroll_x = (hw->x - priv->scroll_x) / (double)sdelta;
+		}
+	    }
+	}
+    }
+    priv->scroll_packet_count = 0;
+}
+
+static int
 HandleScrolling(SynapticsPrivate *priv, struct SynapticsHwState *hw,
 		edge_type edge, Bool finger, struct ScrollData *sd)
 {
     SynapticsSHM *para = priv->synpara;
+    Bool scroll_stop = FALSE;
+    int delay = 1000000000;
 
     sd->left = sd->right = sd->up = sd->down = 0;
 
     /* scroll detection */
     if (finger && !priv->finger_flag) {
+	priv->autoscroll_xspd = 0;
+	priv->autoscroll_yspd = 0;
+	priv->scroll_packet_count = 0;
 	if (para->circular_scrolling) {
 	    if ((para->circular_trigger == 0 && edge) ||
 		(para->circular_trigger == 1 && edge & TOP_EDGE) ||
@@ -1128,6 +1338,7 @@ HandleScrolling(SynapticsPrivate *priv, struct SynapticsHwState *hw,
 		(para->circular_trigger == 7 && edge & LEFT_EDGE) ||
 		(para->circular_trigger == 8 && edge & LEFT_EDGE && edge & TOP_EDGE)) {
 		priv->circ_scroll_on = TRUE;
+		priv->circ_scroll_vert = TRUE;
 		priv->scroll_a = angle(priv, hw->x, hw->y);
 		DBG(7, ErrorF("circular scroll detected on edge\n"));
 	    }
@@ -1148,69 +1359,129 @@ HandleScrolling(SynapticsPrivate *priv, struct SynapticsHwState *hw,
     if (priv->circ_scroll_on && !finger) {
 	/* circular scroll locks in until finger is raised */
 	DBG(7, ErrorF("cicular scroll off\n"));
-	priv->circ_scroll_on = FALSE;
+	scroll_stop = TRUE;
     }
     if (priv->vert_scroll_on && (!(edge & RIGHT_EDGE) || !finger)) {
 	DBG(7, ErrorF("vert edge scroll off\n"));
-	priv->vert_scroll_on = FALSE;
+	scroll_stop = TRUE;
     }
     if (priv->horiz_scroll_on && (!(edge & BOTTOM_EDGE) || !finger)) {
 	DBG(7, ErrorF("horiz edge scroll off\n"));
+	scroll_stop = TRUE;
+    }
+    if (scroll_stop) {
+	start_coasting(priv, hw, edge);
+	priv->circ_scroll_on = FALSE;
+	priv->vert_scroll_on = FALSE;
 	priv->horiz_scroll_on = FALSE;
     }
 
-    /* if hitting a corner (top right or bottom right) while vertical scrolling is active,
-       switch over to circular scrolling smoothly */
+    /* if hitting a corner (top right or bottom right) while vertical scrolling
+       is active, switch over to circular scrolling smoothly */
     if (priv->vert_scroll_on && !priv->horiz_scroll_on && para->circular_scrolling) {
 	if ((edge & RIGHT_EDGE) && (edge & (TOP_EDGE | BOTTOM_EDGE))) {
 	    priv->vert_scroll_on = FALSE;
 	    priv->circ_scroll_on = TRUE;
+	    priv->circ_scroll_vert = TRUE;
+	    priv->scroll_a = angle(priv, hw->x, hw->y);
+	    DBG(7, ErrorF("switching to circular scrolling\n"));
+	}
+    }
+    /* Same treatment for horizontal scrolling */
+    if (priv->horiz_scroll_on && !priv->vert_scroll_on && para->circular_scrolling) {
+	if ((edge & BOTTOM_EDGE) && (edge & (LEFT_EDGE | RIGHT_EDGE))) {
+	    priv->horiz_scroll_on = FALSE;
+	    priv->circ_scroll_on = TRUE;
+	    priv->circ_scroll_vert = FALSE;
 	    priv->scroll_a = angle(priv, hw->x, hw->y);
 	    DBG(7, ErrorF("switching to circular scrolling\n"));
 	}
     }
 
+    if (priv->vert_scroll_on || priv->horiz_scroll_on || priv->circ_scroll_on)
+	priv->scroll_packet_count++;
+
     if (priv->vert_scroll_on) {
 	/* + = down, - = up */
-	while (hw->y - priv->scroll_y > para->scroll_dist_vert) {
-	    sd->down++;
-	    priv->scroll_y += para->scroll_dist_vert;
-	}
-	while (hw->y - priv->scroll_y < -para->scroll_dist_vert) {
-	    sd->up++;
-	    priv->scroll_y -= para->scroll_dist_vert;
-	}
-    }
-    if (priv->circ_scroll_on) {
-	/* + = counter clockwise, - = clockwise */
-	while (diffa(priv->scroll_a, angle(priv, hw->x, hw->y)) > para->scroll_dist_circ) {
-	    sd->up++;
-	    if (sd->up > 1000)
-		break; /* safety */
-	    priv->scroll_a += para->scroll_dist_circ;
-	    if (priv->scroll_a > M_PI)
-		priv->scroll_a -= 2 * M_PI;
-	}
-	while (diffa(priv->scroll_a, angle(priv, hw->x, hw->y)) < -para->scroll_dist_circ) {
-	    sd->down++;
-	    if (sd->down > 1000)
-		break; /* safety */
-	    priv->scroll_a -= para->scroll_dist_circ;
-	    if (priv->scroll_a < -M_PI)
-		priv->scroll_a += 2 * M_PI;
+	int delta = para->scroll_dist_vert;
+	if (delta > 0) {
+	    while (hw->y - priv->scroll_y > delta) {
+		sd->down++;
+		priv->scroll_y += delta;
+	    }
+	    while (hw->y - priv->scroll_y < -delta) {
+		sd->up++;
+		priv->scroll_y -= delta;
+	    }
 	}
     }
     if (priv->horiz_scroll_on) {
 	/* + = right, - = left */
-	while (hw->x - priv->scroll_x > para->scroll_dist_horiz) {
-	    sd->right++;
-	    priv->scroll_x += para->scroll_dist_horiz;
-	}
-	while (hw->x - priv->scroll_x < -para->scroll_dist_horiz) {
-	    sd->left++;
-	    priv->scroll_x -= para->scroll_dist_horiz;
+	int delta = para->scroll_dist_horiz;
+	if (delta > 0) {
+	    while (hw->x - priv->scroll_x > delta) {
+		sd->right++;
+		priv->scroll_x += delta;
+	    }
+	    while (hw->x - priv->scroll_x < -delta) {
+		sd->left++;
+		priv->scroll_x -= delta;
+	    }
 	}
     }
+    if (priv->circ_scroll_on) {
+	/* + = counter clockwise, - = clockwise */
+	double delta = para->scroll_dist_circ;
+	if (delta >= 0.005) {
+	    while (diffa(priv->scroll_a, angle(priv, hw->x, hw->y)) > delta) {
+		if (priv->circ_scroll_vert)
+		    sd->up++;
+		else
+		    sd->right++;
+		priv->scroll_a += delta;
+		if (priv->scroll_a > M_PI)
+		    priv->scroll_a -= 2 * M_PI;
+	    }
+	    while (diffa(priv->scroll_a, angle(priv, hw->x, hw->y)) < -delta) {
+		if (priv->circ_scroll_vert)
+		    sd->down++;
+		else
+		    sd->left++;
+		priv->scroll_a -= delta;
+		if (priv->scroll_a < -M_PI)
+		    priv->scroll_a += 2 * M_PI;
+	    }
+	}
+    }
+
+    if (priv->autoscroll_yspd) {
+	double dtime = (hw->millis - HIST(0).millis) / 1000.0;
+	priv->autoscroll_y += priv->autoscroll_yspd * dtime;
+	delay = MIN(delay, 20);
+	while (priv->autoscroll_y > 1.0) {
+	    sd->down++;
+	    priv->autoscroll_y -= 1.0;
+	}
+	while (priv->autoscroll_y < -1.0) {
+	    sd->up++;
+	    priv->autoscroll_y += 1.0;
+	}
+    }
+    if (priv->autoscroll_xspd) {
+	double dtime = (hw->millis - HIST(0).millis) / 1000.0;
+	priv->autoscroll_x += priv->autoscroll_xspd * dtime;
+	delay = MIN(delay, 20);
+	while (priv->autoscroll_x > 1.0) {
+	    sd->right++;
+	    priv->autoscroll_x -= 1.0;
+	}
+	while (priv->autoscroll_x < -1.0) {
+	    sd->left++;
+	    priv->autoscroll_x += 1.0;
+	}
+    }
+
+    return delay;
 }
 
 /*
@@ -1254,13 +1525,19 @@ HandleState(LocalDevicePtr local, struct SynapticsHwState *hw)
     para->guest_dy = hw->guest_dy;
 
     /* If touchpad is switched off, we skip the whole thing and return delay */
-    if (para->touchpad_off == TRUE)
+    if (para->touchpad_off == 1)
 	return delay;
 
     /* Treat the first two multi buttons as up/down for now. */
     hw->up |= hw->multi[0];
     hw->down |= hw->multi[1];
 
+    if (!para->guestmouse_off) {
+	hw->left |= hw->guest_left;
+	hw->middle |= hw->guest_mid;
+	hw->right |= hw->guest_right;
+    }
+
     /* 3rd button emulation */
     hw->middle |= HandleMidButtonEmulation(priv, hw, &delay);
 
@@ -1302,19 +1579,20 @@ HandleState(LocalDevicePtr local, struct SynapticsHwState *hw)
     if (timeleft > 0)
 	delay = MIN(delay, timeleft);
 
-    HandleScrolling(priv, hw, edge, finger, &scroll);
+    timeleft = HandleScrolling(priv, hw, edge, finger, &scroll);
+    if (timeleft > 0)
+	delay = MIN(delay, timeleft);
 
     timeleft = ComputeDeltas(priv, hw, edge, &dx, &dy);
     delay = MIN(delay, timeleft);
 
-
-    buttons = (((hw->left || hw->guest_left)   ? 0x01 : 0) |
-	       ((hw->middle || hw->guest_mid)  ? 0x02 : 0) |
-	       ((hw->right || hw->guest_right) ? 0x04 : 0) |
-	       (hw->up                         ? 0x08 : 0) |
-	       (hw->down                       ? 0x10 : 0) |
-	       (hw->multi[2]                   ? 0x20 : 0) |
-	       (hw->multi[3]                   ? 0x40 : 0));
+    buttons = ((hw->left     ? 0x01 : 0) |
+	       (hw->middle   ? 0x02 : 0) |
+	       (hw->right    ? 0x04 : 0) |
+	       (hw->up       ? 0x08 : 0) |
+	       (hw->down     ? 0x10 : 0) |
+	       (hw->multi[2] ? 0x20 : 0) |
+	       (hw->multi[3] ? 0x40 : 0));
 
     if (priv->tap_button > 0) {
 	int tap_mask = 1 << (priv->tap_button - 1);
@@ -1446,7 +1724,7 @@ ConvertProc(LocalDevicePtr local,
 }
 
 
-Bool
+static Bool
 QueryHardware(LocalDevicePtr local)
 {
     SynapticsPrivate *priv = (SynapticsPrivate *) local->private;
