@@ -18,6 +18,8 @@ typedef enum {
     MAX_TAP
 } TapEvent;
 
+#define SYN_MAX_BUTTONS 12		    /* Max number of mouse buttons */
+
 struct SynapticsHwInfo {
     unsigned int model_id;		    /* Model-ID */
     unsigned int capabilities;		    /* Capabilities */
@@ -51,6 +53,7 @@ typedef struct _SynapticsSHM
     int tap_move;			    /* max. tapping time and movement in packets and coord. */
     unsigned long tap_time_2;		    /* max. tapping time for double taps */
     unsigned long click_time;		    /* The duration of a single click */
+    Bool fast_taps;			    /* Faster reaction to single taps */
     int emulate_mid_button_time;	    /* Max time between left and right button presses to
 					       emulate a middle button press. */
     int	scroll_dist_vert;		    /* Scrolling distance in absolute coordinates */
@@ -63,7 +66,11 @@ typedef struct _SynapticsSHM
     Bool edge_motion_use_always;	    /* If false, egde motion is used only when dragging */
 
     Bool updown_button_scrolling;	    /* Up/Down-Button scrolling or middle/double-click */
-    Bool touchpad_off;			    /* Switches the touchpad off */
+    int touchpad_off;			    /* Switches the touchpad off
+					     * 0 : Not off
+					     * 1 : Off
+					     * 2 : Only tapping off
+					     */
     Bool guestmouse_off;		    /* Switches the guest mouse off */
     Bool locked_drags;			    /* Enable locked drags */
     int tap_action[MAX_TAP];		    /* Button to report on tap events */
@@ -71,6 +78,10 @@ typedef struct _SynapticsSHM
     double scroll_dist_circ;		    /* Scrolling angle radians */
     int circular_trigger;		    /* Trigger area for circular scrolling */
     Bool circular_pad;			    /* Edge has an oval or circular shape */
+    Bool palm_detect;			    /* Enable Palm Detection */
+    int palm_min_width;			    /* Palm detection width */
+    int palm_min_z;			    /* Palm detection depth */
+    double coasting_speed;		    /* Coasting threshold scrolling speed */
 } SynapticsSHM;
 
 /*
@@ -96,15 +107,10 @@ typedef struct _SynapticsSHM
  *****************************************************************************/
 #define SYNAPTICS_MOVE_HISTORY	5
 
-typedef struct _SynapticsTapRec
-{
-    int x, y;
-    unsigned int millis;
-} SynapticsTapRec;
-
 typedef struct _SynapticsMoveHist
 {
     int x, y;
+    unsigned int millis;
 } SynapticsMoveHistRec;
 
 enum MidButtonEmulation {
@@ -138,15 +144,15 @@ enum TapButtonState {
 
 typedef struct _SynapticsPrivateRec
 {
-    /* shared memory pointer */
-    SynapticsSHM *synpara;
-
+    SynapticsSHM synpara_default;	/* Default parameter settings, read from
+					   the X config file */
+    SynapticsSHM *synpara;		/* Current parameter settings. Will point to
+					   shared memory if shm_config is true */
     struct SynapticsProtocolOperations* proto_ops;
 
     struct SynapticsHwState hwState;
 
-    /* Data read from the touchpad */
-    struct SynapticsHwInfo synhw;
+    struct SynapticsHwInfo synhw;	/* Data read from the touchpad */
     Bool shm_config;			/* True when shared memory area allocated */
 
     OsTimerPtr timer;			/* for up/down-button repeat, tap processing, etc */
@@ -155,6 +161,7 @@ typedef struct _SynapticsPrivateRec
     int fifofd;		 		/* fd for fifo */
 
     SynapticsMoveHistRec move_hist[SYNAPTICS_MOVE_HISTORY]; /* movement history */
+    int hist_index;			/* Last added entry in move_hist[] */
     int largest_valid_x;		/* Largest valid X coordinate seen so far */
     int scroll_y;			/* last y-scroll position */
     int scroll_x;			/* last x-scroll position */
@@ -168,11 +175,18 @@ typedef struct _SynapticsPrivateRec
     int tap_max_fingers;		/* Max number of fingers seen since entering start state */
     int tap_button;			/* Which button started the tap processing */
     enum TapButtonState tap_button_state; /* Current tap action */
-    SynapticsTapRec touch_on;		/* data when the touchpad is touched/released */
+    SynapticsMoveHistRec touch_on;	/* data when the touchpad is touched/released */
 
     Bool vert_scroll_on;		/* scrolling flag */
     Bool horiz_scroll_on;		/* scrolling flag */
     Bool circ_scroll_on;		/* scrolling flag */
+    Bool circ_scroll_vert;		/* True: Generate vertical scroll events
+					   False: Generate horizontal events */
+    double autoscroll_xspd;		/* Horizontal coasting speed */
+    double autoscroll_yspd;		/* Vertical coasting speed */
+    double autoscroll_x;		/* Accumulated horizontal coasting scroll */
+    double autoscroll_y;		/* Accumulated vertical coasting scroll */
+    int scroll_packet_count;		/* Scroll duration */
     double frac_x, frac_y;		/* absolute -> relative fraction */
     enum MidButtonEmulation mid_emu_state;	/* emulated 3rd button */
     int repeatButtons;			/* buttons for repeat */
@@ -186,8 +200,6 @@ typedef struct _SynapticsPrivateRec
 } SynapticsPrivate;
 
 
-Bool QueryHardware(LocalDevicePtr);
-
 #endif /* SYNAPTICS_PRIVATE */
 
 
